from __future__ import annotations

import os
import re
from typing import Optional

import rdflib
from hdbcli import dbapi
from rdflib.plugins.sparql import prepareQuery


class HanaRdfGraph:
    """
    SAP HANA CLOUD Knowledge Graph Engine Wrapper

    This class connects to a SAP HANA Graph SPARQL endpoint, executes queries,
    and loads or generates ontology/schema data via one of four methods:

    1. ontology_query: Provide a SPARQL CONSTRUCT query to extract the schema
       directly from the graph.
    2. ontology_uri: Specify a remote ontology graph URI; the schema is loaded
       via a default CONSTRUCT that copies all triples.
    3. ontology_local_file: Load the schema from a local RDF file (e.g., Turtle,
       RDF/XML) on disk.
    4. auto_extract_ontology: When enabled (and no other source given), run a
       built-in generic CONSTRUCT query that reverse-engineers classes, properties,
       domains, and ranges from your instance data.

    Args:
        connection (dbapi.Connection): A HANA database connection instance.
        ontology_query (Optional[str]): A SPARQL CONSTRUCT query to load the schema.
        ontology_uri (Optional[str]): The URI of the ontology graph
            containing the RDF schema.
        ontology_local_file (Optional[str]): Path to a local ontology file to load.
        ontology_local_file_format (Optional[str]): RDF format of the local file
            (e.g., 'turtle').
        graph_uri (Optional[str]): The URI of the target graph; uses the DEFAULT graph if graph_uri in (None, "", "DEFAULT").
        auto_extract_ontology (bool): If True and no schema source provided,
            automatically extract a generic ontology via SPARQL.

    Example:
        from hdbcli import dbapi
        # Establish a database connection
        connection = dbapi.connect(
            address="<hostname>",
            port=30015,
            user="<username>",
            password="<password>"
        )
        rdf_graph = HanaRdfGraph(
            connection=connection,
            graph_uri="http://example.com/graph",
            ontology_uri="http://example.com/ontology"
        )
        # Execute a SPARQL query:
        sparql_query = "SELECT ?s ?p ?o WHERE { ?s ?p ?o . }"
        response = rdf_graph.query(sparql_query, inject_from_clause=True)
        print(response)

    *Security note*: Make sure that the database connection uses credentials
        that are narrowly-scoped to only include necessary permissions.
        Failure to do so may result in data corruption or loss, since the calling
        code may attempt commands that would result in deletion, mutation
        of data if appropriately prompted or reading sensitive data if such
        data is present in the database.
        The best way to guard against such negative outcomes is to (as appropriate)
        limit the permissions granted to the credentials used with this tool.

        See https://python.langchain.com/docs/security for more information.
    """

    def __init__(
        self,
        connection: dbapi.Connection,
        graph_uri: Optional[
            str
        ] = "",  # use default graph if graph_uri in (None, "", "DEFAULT")
        ontology_query: Optional[str] = None,
        ontology_uri: Optional[str] = None,
        ontology_local_file: Optional[str] = None,
        ontology_local_file_format: Optional[str] = None,
        auto_extract_ontology: bool = False,
    ) -> None:
        self.connection = connection

        # If graph_uri is None, empty, or "DEFAULT", use the DEFAULT graph.
        if not graph_uri or graph_uri.upper() == "DEFAULT":
            self.from_clause = "FROM DEFAULT"
        else:
            self.from_clause = f"FROM <{graph_uri}>"

        self.refresh_schema(
            ontology_query,
            ontology_uri,
            ontology_local_file,
            ontology_local_file_format,
            auto_extract_ontology,
        )

    def inject_from_clause(self, query: str) -> str:
        """
        Injects a FROM clause into the SPARQL query if one is not already present..

        If self.graph_uri is provided, it inserts FROM <graph_uri>.

        Args:
            query: The SPARQL query string.

        Returns:
            The modified SPARQL query with the appropriate FROM clause.

        Raises:
            ValueError: If the query does not contain a 'WHERE' clause.
        """

        # Check if a FROM clause is already present.
        from_pattern = re.compile(r"\bFROM\b", flags=re.IGNORECASE)
        if from_pattern.search(query):
            # FROM clause already exists, return query unchanged.
            return query

        # Use regex to match the first occurrence of 'WHERE' with word boundaries
        pattern = re.compile(r"\bWHERE\b", flags=re.IGNORECASE)
        match = pattern.search(query)
        if match:
            index = match.start()
            # Insert the FROM clause before the matched WHERE clause.
            query = query[:index] + f"\n{self.from_clause}\n" + query[index:]
        else:
            raise ValueError("The SPARQL query does not contain a 'WHERE' clause.")

        return query

    def query(
        self,
        query: str,
        content_type: Optional[str] = None,
        inject_from_clause: bool = True,  # If True , inject a FROM clause into query.
    ) -> str:
        """Executes SPARQL query and returns response as a string."""

        if content_type is None:
            content_type = "application/sparql-results+csv"

        request_headers = (
            f"Accept: {content_type}\r\nContent-Type: application/sparql-query"
        )

        if inject_from_clause:
            query = self.inject_from_clause(query)

        cursor = self.connection.cursor()
        try:
            result = cursor.callproc(
                "SYS.SPARQL_EXECUTE", (query, request_headers, "?", "?")
            )
            response = result[2]
        except dbapi.Error as db_error:
            raise RuntimeError(
                f'The database query "{query}" failed: '
                f'{db_error.errortext.split("; Server Connection")[0]}'
            )

        finally:
            cursor.close()

        return response

    def _load_ontology_schema_graph_from_query(self, ontology_query) -> rdflib.Graph:
        """
        Load an ontology schema by executing a SPARQL CONSTRUCT query.
        """
        self._validate_construct_query(ontology_query)

        response = self.query(ontology_query, content_type="", inject_from_clause=False)

        graph = rdflib.Graph()

        # Parse the string into the graph
        graph.parse(data=response, format="turtle")

        return graph

    @staticmethod
    def _load_ontology_schema_from_file(local_file: str, local_file_format: str = None):  # type: ignore[no-untyped-def, assignment]
        """
        Parse the ontology schema statements from the provided file
        """
        if not os.path.exists(local_file):
            raise FileNotFoundError(f"File {local_file} does not exist.")
        if not os.access(local_file, os.R_OK):
            raise PermissionError(f"Read permission for {local_file} is restricted")
        graph = rdflib.Graph()
        try:
            graph.parse(local_file, format=local_file_format)
        except Exception as e:
            raise ValueError(f"Invalid file format for {local_file} : ", e)
        return graph

    def refresh_schema(
        self,
        ontology_query: Optional[str] = None,
        ontology_uri: Optional[str] = None,
        ontology_local_file: Optional[str] = None,
        ontology_local_file_format: Optional[str] = None,
        auto_extract_ontology: bool = False,
    ) -> None:
        """
        Load or generate the graph's ontology schema.

        Args:
            ontology_query: SPARQL CONSTRUCT to load schema.
            ontology_uri: URI of schema graph to load.
            ontology_local_file: Local file path for schema.
            ontology_local_file_format: Format of local file.
            auto_extract_ontology: If True and no other source, use generic extractor.

        Raises:
            ValueError: If multiple or no schema sources are specified.
        """
        # Count provided schema sources
        schema_sources = sum(
            1
            for source in [ontology_query, ontology_uri, ontology_local_file]
            if source is not None
        )

        if schema_sources == 0 and auto_extract_ontology:
            ontology_query = self._get_generic_ontology_query()
            schema_sources = 1

        if schema_sources > 1:
            raise ValueError(
                "Multiple ontology/schema sources provided. Use only one of: "
                "ontology_query, ontology_uri, or ontology_local_file."
            )
        elif schema_sources == 0:
            raise ValueError(
                "No ontology/schema sources provided. Use only one of: "
                "ontology_query, ontology_uri, or ontology_local_file."
            )

        if ontology_local_file:
            ontology_schema_graph = self._load_ontology_schema_from_file(
                ontology_local_file,
                ontology_local_file_format,  # type: ignore[arg-type]
            )
        else:
            if ontology_uri:
                ontology_query = (
                    f"CONSTRUCT {{?s ?p ?o}} FROM <{ontology_uri}> WHERE"
                    + "{?s ?p ?o .}"
                )
            ontology_schema_graph = self._load_ontology_schema_graph_from_query(
                ontology_query
            )

        self.schema = ontology_schema_graph

    @staticmethod
    def _validate_construct_query(construct_query: str) -> None:
        """
        Validate the query is a valid SPARQL CONSTRUCT query.

        Args:
            construct_query: The SPARQL query string to validate.

        Raises:
            TypeError: If query is not a string.
            ValueError: If query is not a valid CONSTRUCT query.
        """
        if not isinstance(construct_query, str):
            raise TypeError("Schema query must be provided as string.")

        # We are using prepareQuery from rdflib.plugins.sparql to parse the query.
        # HANA deviates from the standard SPARQL specification where FROM DEFAULT is a valid clause.
        # To handle this, we remove 'FROM DEFAULT' before parsing.
        # In the standard specification, when no FROM clause is given, the default graph is used. 
        query_without_from_default = re.sub(r'\bfrom\s+default\b', '', construct_query, flags=re.IGNORECASE)
        parsed_query = prepareQuery(query_without_from_default)
        if parsed_query.algebra.name != "ConstructQuery":
            raise ValueError(
                "Invalid query type. Only CONSTRUCT queries are supported for schema."
            )

    def _get_generic_ontology_query(self):
        """
        Return a generic SPARQL CONSTRUCT that extracts
            a minimal OWL schema from the graph given by self.graph_uri.

        Returns:
            A SPARQL CONSTRUCT query string.
        """
        ontology_query = f"""
        PREFIX owl: <http://www.w3.org/2002/07/owl#>
        PREFIX rdf: <http://www.w3.org/1999/02/22-rdf-syntax-ns#>
        PREFIX rdfs: <http://www.w3.org/2000/01/rdf-schema#>
        PREFIX xsd: <http://www.w3.org/2001/XMLSchema#>
        CONSTRUCT {{ ?cls rdf:type owl:Class . ?cls rdfs:label ?clsLabel . ?rel rdf:type ?propertyType . ?rel rdfs:label ?relLabel . ?rel rdfs:domain ?domain . ?rel rdfs:range ?range .}}
        {self.from_clause}
        WHERE {{ # get properties
            {{SELECT DISTINCT ?domain ?rel ?relLabel ?propertyType ?range
            WHERE {{
                ?subj ?rel ?obj .
                ?subj a ?domain .
                OPTIONAL{{?obj a ?rangeClass .}}
                FILTER(?rel != rdf:type) #I think we should remove type
                BIND(IF(isIRI(?obj) = true, owl:ObjectProperty, owl:DatatypeProperty) AS ?propertyType)
                BIND(COALESCE(?rangeClass, DATATYPE(?obj)) AS ?range)
                BIND(STR(?rel) AS ?uriStr)       # Convert URI to string
                BIND(REPLACE(?uriStr, "^.*[/#]", "") AS ?relLabel)
            }}}}
            UNION {{ # get classes
                SELECT DISTINCT ?cls ?clsLabel
                WHERE {{
                    ?instance a/rdfs:subClassOf* ?cls .
                    FILTER (isIRI(?cls)) .
                    BIND(STR(?cls) AS ?uriStr)       # Convert URI to string
                    BIND(REPLACE(?uriStr, "^.*[/#]", "") AS ?clsLabel)
                }}
            }}
        }}
        """
        return ontology_query

    @property
    def get_schema(self) -> rdflib.Graph:
        """
        Return the schema of the graph as an rdflib.Graph object.
        
        To get the schema in turtle format, use: graph.get_schema.serialize(format="turtle")
        """
        return self.schema



from langchain_core.prompts.prompt import PromptTemplate

SPARQL_GENERATION_SELECT_TEMPLATE = """
Given the ontology below, create a SPARQL SELECT query from the user prompt.
Generate only SELECT queries - do not generate INSERT, UPDATE, DELETE, CREATE, DROP, or any other modification queries.
Enclose literals in double quotes. Note that the graph is directed. Edges go from the domain to the range.
If an RDFS label exists for a class or a property, always retrieve the label.
Use only the entity types and properties provided in the ontology.
Do not use any entity types and properties that are not explicitly provided.
Include all necessary prefixes.
For instance, to find all actors of the movie "Blade Runner", the following SELECT query in fenced code blocks would be suitable:
```sparql
PREFIX kg: <http://kg.demo.sap.com/>
PREFIX rdf: <http://www.w3.org/1999/02/22-rdf-syntax-ns#>
PREFIX rdfs: <http://www.w3.org/2000/01/rdf-schema#>
SELECT *
WHERE {{
    ?movie rdf:type kg:Film .
    ?movie kg:title ?movieTitle .
    ?actor kg:acted_in ?movie .
    ?actor rdfs:label ?actorLabel .
    FILTER(?movieTitle = "Blade Runner")
}}
```
<ontology>
{schema}
</ontology>
Do not respond to any questions that ask for anything else than for you to construct a SPARQL SELECT query.
Always enclose your SPARQL SELECT query response in fenced code blocks using ```sparql at the beginning and ``` at the end.
Do not include any text except the SPARQL SELECT query generated within the fenced code blocks.
Only generate read-only SELECT queries. Never generate queries that modify data (INSERT, UPDATE, DELETE, CREATE, DROP, etc.).
Please pay attention to providing the subject, predicate, and object in the correct order.
Ensure that every variable referenced in any clause (such as SELECT, ORDER BY, GROUP BY, etc.) is explicitly defined in the WHERE clause, either by being used as a subject, predicate, or object in a triple pattern, or through a BIND statement.
Do not include any variables in those clauses unless they are defined in the WHERE clause.

The question is:
{prompt}"""
SPARQL_GENERATION_SELECT_PROMPT = PromptTemplate(
    input_variables=["schema", "prompt"], template=SPARQL_GENERATION_SELECT_TEMPLATE
)
SPARQL_QA_TEMPLATE = """Task: Generate a natural language response from the results of a SPARQL query.
You are an assistant that creates well-written and human understandable answers.
The information part contains the information provided, which you can use to construct an answer.
The information provided is authoritative, you must never doubt it or try to use your internal knowledge to correct it.
Make your response sound like the information is coming from an AI assistant, but don't add any information. 
Don't use internal knowledge to answer the question, just say you don't know if no information is available.
Information:
{context}

Question: {prompt}
Helpful Answer:"""
SPARQL_QA_PROMPT = PromptTemplate(
    input_variables=["context", "prompt"], template=SPARQL_QA_TEMPLATE
)



"""Question answering over a SAP HANA graph using SPARQL."""

from __future__ import annotations

from typing import Any, Dict, List, Optional

from langchain.chains.base import Chain
from langchain.chains.llm import LLMChain
from langchain_core.callbacks.manager import CallbackManagerForChainRun
from langchain_core.language_models import BaseLanguageModel
from langchain_core.prompts.base import BasePromptTemplate
from pydantic import Field

from langchain_hana.graphs import HanaRdfGraph

from .prompts import (
    SPARQL_GENERATION_SELECT_PROMPT,
    SPARQL_QA_PROMPT,
)


class HanaSparqlQAChain(Chain):
    """Chain for question-answering against a SAP HANA CLOUD Knowledge Graph Engine
    by generating SPARQL statements.

    Example:
        chain = HanaSparqlQAChain.from_llm(
            llm=llm,
            verbose=True,
            allow_dangerous_requests=True,
            graph=graph)
        response = chain.invoke(query)

    *Security note*: Make sure that the database connection uses credentials
        that are narrowly-scoped to only include necessary permissions.
        Failure to do so may result in data corruption or loss, since the calling
        code may attempt commands that would result in deletion, mutation
        of data if appropriately prompted or reading sensitive data if such
        data is present in the database.
        The best way to guard against such negative outcomes is to (as appropriate)
        limit the permissions granted to the credentials used with this tool.

        See https://python.langchain.com/docs/security for more information.
    """

    graph: HanaRdfGraph = Field(exclude=True)
    sparql_generation_chain: LLMChain
    qa_chain: LLMChain
    input_key: str = "query"  #: :meta private:
    output_key: str = "result"  #: :meta private:

    allow_dangerous_requests: bool = False

    def __init__(self, **kwargs: Any) -> None:
        """Initialize the chain."""
        super().__init__(**kwargs)
        if self.allow_dangerous_requests is not True:
            raise ValueError(
                "In order to use this chain, you must acknowledge that it can make "
                "dangerous requests by setting `allow_dangerous_requests` to `True`."
                "You must narrowly scope the permissions of the database connection "
                "to only include necessary permissions. Failure to do so may result "
                "in data corruption or loss or reading sensitive data if such data is "
                "present in the database."
                "Only use this chain if you understand the risks and have taken the "
                "necessary precautions. "
                "See https://python.langchain.com/docs/security for more information."
            )

    @property
    def input_keys(self) -> List[str]:
        return [self.input_key]

    @property
    def output_keys(self) -> List[str]:
        return [self.output_key]

    @classmethod
    def from_llm(
        cls,
        llm: BaseLanguageModel,
        *,
        sparql_generation_prompt: BasePromptTemplate = SPARQL_GENERATION_SELECT_PROMPT,
        qa_prompt: BasePromptTemplate = SPARQL_QA_PROMPT,
        **kwargs: Any,
    ) -> HanaSparqlQAChain:
        sparql_generation_chain = LLMChain(llm=llm, prompt=sparql_generation_prompt)
        qa_chain = LLMChain(llm=llm, prompt=qa_prompt)
        return cls(
            qa_chain=qa_chain,
            sparql_generation_chain=sparql_generation_chain,
            **kwargs,
        )

    @staticmethod
    def extract_sparql(query: str) -> str:
        """Extract SPARQL code from a text.

        Args:
            query: Text to extract SPARQL code from.

        Returns:
            SPARQL code extracted from the text.
        """
        query = query.strip()
        querytoks = query.split("```")
        if len(querytoks) == 3:
            query = querytoks[1]
            if query.lower().startswith("sparql"):
                query = query[6:]
        elif query.startswith("<sparql>") and query.endswith("</sparql>"):
            query = query[8:-9]
        return query

    def _ensure_common_prefixes(self, query: str) -> str:
        """
        Ensure common prefixes (rdf, rdfs, owl, xsd) are declared if used in the query.
        """
        common = {
            "rdf": "http://www.w3.org/1999/02/22-rdf-syntax-ns#",
            "rdfs": "http://www.w3.org/2000/01/rdf-schema#",
            "owl": "http://www.w3.org/2002/07/owl#",
            "xsd": "http://www.w3.org/2001/XMLSchema#",
        }
        # detect existing prefix declarations
        present = set()
        for line in query.splitlines():
            line_strip = line.strip()
            if line_strip.upper().startswith("PREFIX "):
                parts = line_strip.split()
                if len(parts) >= 2 and parts[1].endswith(":"):
                    present.add(parts[1][:-1])
        # build missing declarations
        missing = [(p, uri) for p, uri in common.items() if p not in present]
        if missing:
            prefix_lines = ""
            for p, uri in missing:
                if p in query:
                    prefix_lines += f"PREFIX {p}: <{uri}>\n"
            query = prefix_lines + query
        return query

    def _call(
        self,
        inputs: Dict[str, Any],
        run_manager: Optional[CallbackManagerForChainRun] = None,
        is_insert=False,
    ) -> Dict[str, str]:
        "Generate SPARQL query, use it to look up in the graph and answer the question."
        _run_manager = run_manager or CallbackManagerForChainRun.get_noop_manager()
        callbacks = _run_manager.get_child()

        # Extract user question
        question = inputs[self.input_key]

        # Generate SPARQL query from the question and schema
        sparql_result = self.sparql_generation_chain.invoke(
            {
                "prompt": question, 
                "schema": self.graph.get_schema.serialize(format="turtle")
            }, 
            callbacks=callbacks
        )
        # Extract the generated SPARQL string from the result dictionary
        generated_sparql = sparql_result[self.sparql_generation_chain.output_key]

        # Log the generated SPARQL
        _run_manager.on_text("Generated SPARQL:", end="\n", verbose=self.verbose)
        _run_manager.on_text(
            generated_sparql, color="green", end="\n", verbose=self.verbose
        )

        # Extract the SPARQL code from the generated text and inject the from clause
        generated_sparql = self.extract_sparql(generated_sparql)
        generated_sparql = self.graph.inject_from_clause(generated_sparql)
        generated_sparql = self._ensure_common_prefixes(generated_sparql)
        _run_manager.on_text("Final SPARQL:", end="\n", verbose=self.verbose)
        _run_manager.on_text(
            generated_sparql, color="yellow", end="\n", verbose=self.verbose
        )

        # Execute the generated SPARQL query against the graph
        context = self.graph.query(generated_sparql, inject_from_clause=False)

        # Log the full context (SPARQL results)
        _run_manager.on_text("Full Context:", end="\n", verbose=self.verbose)
        _run_manager.on_text(
            str(context), color="green", end="\n", verbose=self.verbose
        )

        # Pass the question and query results into the QA chain
        qa_chain_result = self.qa_chain.invoke(
            {"prompt": question, "context": context}, callbacks=callbacks
        )
        # Extract the final answer from the result dictionary
        result = qa_chain_result[self.qa_chain.output_key]

        # Return the final answer
        return {self.output_key: result}



"""SAP HANA Cloud Vector Engine"""

from __future__ import annotations

import json
import logging
import re
import struct
from typing import (
    Any,
    Callable,
    Iterable,
    Optional,
    Pattern,
    Type,
)

import numpy as np
from hdbcli import dbapi  # type: ignore
from langchain_core.documents import Document
from langchain_core.embeddings import Embeddings
from langchain_core.runnables.config import run_in_executor
from langchain_core.vectorstores import VectorStore
from langchain_core.vectorstores.utils import maximal_marginal_relevance

from langchain_hana.embeddings import HanaInternalEmbeddings
from langchain_hana.vectorstores.create_where_clause import (
    CONTAINS_OPERATOR,
    LOGICAL_OPERATORS_TO_SQL,
    CreateWhereClause,
)
from langchain_hana.utils import DistanceStrategy, _validate_k, _validate_k_and_fetch_k

logger = logging.getLogger(__name__)

HANA_DISTANCE_FUNCTION: dict = {
    DistanceStrategy.COSINE: ("COSINE_SIMILARITY", "DESC"),
    DistanceStrategy.EUCLIDEAN_DISTANCE: ("L2DISTANCE", "ASC"),
}

VECTOR_COLUMN_SQL_TYPES = ["REAL_VECTOR", "HALF_VECTOR"]

INTERMEDIATE_TABLE_NAME = "intermediate_result"

default_distance_strategy = DistanceStrategy.COSINE
default_table_name: str = "EMBEDDINGS"
default_content_column: str = "VEC_TEXT"
default_metadata_column: str = "VEC_META"
default_vector_column: str = "VEC_VECTOR"
default_vector_column_length: int = -1  # -1 means dynamic length
default_vector_column_type: str = "REAL_VECTOR"


class HanaDB(VectorStore):
    """SAP HANA Cloud Vector Engine

    The prerequisite for using this class is the installation of the ``hdbcli``
    Python package.

    The HanaDB vectorstore can be created by providing an embedding function and
    an existing database connection. Optionally, the names of the table and the
    columns to use.
    """

    def __init__(
        self,
        connection: dbapi.Connection,
        embedding: Embeddings,
        distance_strategy: DistanceStrategy = default_distance_strategy,
        table_name: str = default_table_name,
        content_column: str = default_content_column,
        metadata_column: str = default_metadata_column,
        vector_column: str = default_vector_column,
        vector_column_length: int = default_vector_column_length,
        vector_column_type: str = default_vector_column_type,
        *,
        specific_metadata_columns: Optional[list[str]] = None,
    ):
        valid_distance = False
        for key in HANA_DISTANCE_FUNCTION.keys():
            if key is distance_strategy:
                valid_distance = True
        if not valid_distance:
            raise ValueError(
                "Unsupported distance_strategy: {}".format(distance_strategy)
            )

        self.connection = connection
        self.distance_strategy = distance_strategy
        self.table_name = HanaDB._sanitize_name(table_name)
        self.content_column = HanaDB._sanitize_name(content_column)
        self.metadata_column = HanaDB._sanitize_name(metadata_column)
        self.vector_column = HanaDB._sanitize_name(vector_column)
        self.vector_column_length = HanaDB._sanitize_int(vector_column_length)
        self.vector_column_type = HanaDB._sanitize_vector_column_type(
            vector_column_type, connection
        )
        self.specific_metadata_columns = HanaDB._sanitize_specific_metadata_columns(
            specific_metadata_columns or []
        )

        # Configure the embedding (internal or external)
        self.embedding: Embeddings
        self.use_internal_embeddings: bool = False
        self.internal_embedding_model_id: str = ""
        self.internal_embedding_remote_source: str = ""
        self.set_embedding(embedding)

        # Initialize the table if it doesn't exist
        self._initialize_table()

    def set_embedding(self, embedding: Embeddings) -> None:
        """
        Use this method if you need to change the embedding instance
        """
        self.embedding = embedding
        # Decide whether to use internal or external embeddings
        if isinstance(embedding, HanaInternalEmbeddings):
            # Internal embeddings
            self.use_internal_embeddings = True
            self.internal_embedding_model_id = embedding.get_model_id()
            self.internal_embedding_remote_source = embedding.get_remote_source()
            self._validate_internal_embedding_function()
        else:
            # External embeddings
            self.use_internal_embeddings = False
            self.internal_embedding_model_id = ""

    def _initialize_table(self) -> None:
        """Create the table if it doesn't exist and validate columns."""
        if not self._table_exists(self.table_name):
            sql_str = (
                f'CREATE TABLE "{self.table_name}"('
                f'"{self.content_column}" NCLOB, '
                f'"{self.metadata_column}" NCLOB, '
                f'"{self.vector_column}" {self.vector_column_type} '
            )
            if self.vector_column_length in [-1, 0]:
                sql_str += ");"
            else:
                sql_str += f"({self.vector_column_length}));"

            try:
                cur = self.connection.cursor()
                cur.execute(sql_str)
            finally:
                cur.close()

        # Check if the needed columns exist and have the correct type
        self._check_column(self.table_name, self.content_column, ["NCLOB", "NVARCHAR"])
        self._check_column(self.table_name, self.metadata_column, ["NCLOB", "NVARCHAR"])
        self._check_column(
            self.table_name,
            self.vector_column,
            [self.vector_column_type],
            self.vector_column_length,
        )
        for column_name in self.specific_metadata_columns:
            self._check_column(self.table_name, column_name)

    def _table_exists(self, table_name) -> bool:  # type: ignore[no-untyped-def]
        sql_str = (
            "SELECT COUNT(*) FROM SYS.TABLES WHERE SCHEMA_NAME = CURRENT_SCHEMA"
            " AND TABLE_NAME = ?"
        )
        try:
            cur = self.connection.cursor()
            cur.execute(sql_str, (table_name))
            if cur.has_result_set():
                rows = cur.fetchall()
                if rows[0][0] == 1:
                    return True
        finally:
            cur.close()
        return False

    def _check_column(  # type: ignore[no-untyped-def]
        self, table_name, column_name, column_type=None, column_length=None
    ):
        sql_str = (
            "SELECT DATA_TYPE_NAME, LENGTH FROM SYS.TABLE_COLUMNS WHERE "
            "SCHEMA_NAME = CURRENT_SCHEMA "
            "AND TABLE_NAME = ? AND COLUMN_NAME = ?"
        )
        try:
            cur = self.connection.cursor()
            cur.execute(sql_str, (table_name, column_name))
            if cur.has_result_set():
                rows = cur.fetchall()
                if len(rows) == 0:
                    raise AttributeError(f"Column {column_name} does not exist")
                # Check data type
                if column_type:
                    if rows[0][0] not in column_type:
                        raise AttributeError(
                            f"Column {column_name} has the wrong type: {rows[0][0]}"
                        )
                # Check length, if parameter was provided
                # Length can either be -1 (QRC01+02-24) or 0 (QRC03-24 onwards)
                # to indicate no length constraint being present.
                if column_length is not None and column_length > 0:
                    if rows[0][1] != column_length:
                        raise AttributeError(
                            f"Column {column_name} has the wrong length: {rows[0][1]} "
                            f"expected: {column_length}"
                        )
            else:
                raise AttributeError(f"Column {column_name} does not exist")
        finally:
            cur.close()

    def _generate_vector_embedding_sql_and_params(self, text: str, type: str) -> str:
        """Generate the VECTOR_EMBEDDING SQL expression and parameters."""
        if type != "QUERY" and type != "DOCUMENT":
            raise ValueError("type must be either 'QUERY' or 'DOCUMENT'")
        
        vector_embedding_params = [text, self.internal_embedding_model_id]
        if  not self.internal_embedding_remote_source:
            vector_embedding_sql = (
                f"""VECTOR_EMBEDDING(?, '{type}', ?)"""
            )
        else:
            vector_embedding_sql = (
                f"""VECTOR_EMBEDDING(?, '{type}', ?, "{self.internal_embedding_remote_source}")"""
            )

        return vector_embedding_sql, vector_embedding_params


    def _validate_internal_embedding_function(self) -> None:
        """
        Ping the database to check if the in-database embedding function
            exists and works.
        Raises:
            ValueError: If the embedding function does not exist or fails.
        """
        if self.internal_embedding_model_id is None:
            raise ValueError("Internal embedding model id can't be none!")
        vector_embedding_sql, vector_embedding_params = self._generate_vector_embedding_sql_and_params('test', 'QUERY')
        cur = self.connection.cursor()
        try:
            # Test the VECTOR_EMBEDDING function by executing a simple query
            cur.execute(
                (
                    "SELECT COUNT(TO_NVARCHAR("
                    f"{vector_embedding_sql})) "
                    ' AS "CNT" FROM sys.DUMMY;'
                ), vector_embedding_params
            )
        finally:
            cur.close()

    @property
    def embeddings(self) -> Embeddings:
        return self.embedding

    @staticmethod
    def _sanitize_name(input_str: str) -> str:  # type: ignore[misc]
        # Remove characters that are not alphanumeric or underscores
        return re.sub(r"[^a-zA-Z0-9_]", "", input_str)

    @staticmethod
    def _sanitize_int(input_int: any) -> int:  # type: ignore[valid-type]
        value = int(str(input_int))
        if value < -1:
            raise ValueError(f"Value ({value}) must not be smaller than -1")
        return int(str(input_int))

    @staticmethod
    def _sanitize_list_float(embedding: list[float]) -> list[float]:
        for value in embedding:
            if not isinstance(value, float):
                raise ValueError(f"Value ({value}) does not have type float")
        return embedding

    # Compile pattern only once, for better performance
    _compiled_pattern: Pattern = re.compile("^[_a-zA-Z][_a-zA-Z0-9]*$")

    @staticmethod
    def _sanitize_metadata_keys(metadata: dict) -> dict:
        for key in metadata.keys():
            if not HanaDB._compiled_pattern.match(key):
                raise ValueError(f"Invalid metadata key {key}")

        return metadata

    @staticmethod
    def _sanitize_specific_metadata_columns(
        specific_metadata_columns: list[str],
    ) -> list[str]:
        metadata_columns = []
        for c in specific_metadata_columns:
            sanitized_name = HanaDB._sanitize_name(c)
            metadata_columns.append(sanitized_name)
        return metadata_columns

    @staticmethod
    def _sanitize_vector_column_type(
        vector_column_type: str, connection: dbapi.Connection
    ) -> str:
        vector_column_type_upper = vector_column_type.upper()
        if vector_column_type_upper not in VECTOR_COLUMN_SQL_TYPES:
            raise ValueError(
                f"Unsupported vector_column_type: {vector_column_type}. "
                f"Must be one of {', '.join(VECTOR_COLUMN_SQL_TYPES)}"
            )
        HanaDB._validate_datatype_support(connection, vector_column_type_upper)
        return vector_column_type_upper

    @staticmethod
    def _get_min_supported_version(datatype: str) -> str:
        if datatype == "HALF_VECTOR":
            return "2025.15 (QRC 2/2025)"
        elif datatype == "REAL_VECTOR":
            return "2024.2 (QRC 1/2024)"
        else:
            raise ValueError(f"Unknown datatype: '{datatype}'")

    @staticmethod
    def _get_instance_version(connection: dbapi.Connection) -> Optional[str]:
        cursor = connection.cursor()
        try:
            cursor.execute("SELECT CLOUD_VERSION FROM SYS.M_DATABASE;")
            result = cursor.fetchone()
            return result[0]
        except dbapi.Error:
            return None
        finally:
            cursor.close()

    @staticmethod
    def _get_available_datatypes(connection: dbapi.Connection) -> set:
        cur = connection.cursor()
        try:
            cur.execute("SELECT TYPE_NAME FROM SYS.DATA_TYPES")
            if cur.has_result_set():
                rows = cur.fetchall()
                available_types = {row[0] for row in rows}
                return available_types
            raise ValueError("No data types returned by the database.")
        finally:
            cur.close()

    @staticmethod
    def _validate_datatype_support(connection: dbapi.Connection, datatype: str) -> bool:
        if datatype in HanaDB._get_available_datatypes(connection):
            return True

        # Get instance version, but don't include it in error if retrieval fails
        error_message = f"'{datatype}' is not available on this HANA instance.\n"

        # Only include instance version line if it was successfully retrieved
        instance_version = HanaDB._get_instance_version(connection)
        if instance_version:
            error_message += f"Instance version: {instance_version}\n"

        min_instance_version = HanaDB._get_min_supported_version(datatype)
        error_message += f"Minimum required instance version: {min_instance_version}"

        raise ValueError(error_message)

    def _serialize_binary_format(self, values: list[float]) -> bytes:
        # Converts a list of floats into binary format
        if self.vector_column_type == "HALF_VECTOR":
            # 2-byte half-precision float serialization
            return struct.pack(f"<I{len(values)}e", len(values), *values)
        elif self.vector_column_type == "REAL_VECTOR":
            # 4-byte float serialization (standard FVECS format)
            return struct.pack(f"<I{len(values)}f", len(values), *values)
        else:
            raise ValueError(
                f"Unsupported vector column type: {self.vector_column_type}"
            )

    def _handle_vector_output_type(self, vector: Any) -> list[float]:
        # Handles different vector output types from HANA DB
        if isinstance(vector, memoryview):
            return self._deserialize_binary_format(vector)
        
        if isinstance(vector, tuple):
            return list(vector)
        
        if isinstance(vector, list):
            return vector

        raise ValueError("Unsupported vector format returned from database.")

    def _deserialize_binary_format(self, fvecs: memoryview) -> list[float]:
        # Extracts a list of floats from binary format
        dim = struct.unpack_from("<I", fvecs, 0)[0]
        if self.vector_column_type == "HALF_VECTOR":
            # 2-byte half-precision float deserialization
            return list(struct.unpack_from(f"<{dim}e", fvecs, 4))
        elif self.vector_column_type == "REAL_VECTOR":
            # 4-byte float deserialization (standard FVECS format)
            return list(struct.unpack_from(f"<{dim}f", fvecs, 4))
        else:
            raise ValueError(
                f"Unsupported vector column type: {self.vector_column_type}"
            )

    def _split_off_special_metadata(self, metadata: dict) -> tuple[dict, list]:
        # Use provided values by default or fallback
        special_metadata = []

        if not metadata:
            return {}, []

        for column_name in self.specific_metadata_columns:
            special_metadata.append(metadata.get(column_name, None))

        return metadata, special_metadata

    def _convert_to_target_vector_type(self, expr: str) -> str:
        """
        Converts a vector expression to the target vector column type.

        Applies the appropriate vector conversion function
            (TO_REAL_VECTOR or TO_HALF_VECTOR) to the provided
            expression based on the configured vector_column_type.

        Args:
            expr (str): a vector expression

        Returns:
            str: The expression wrapped with the appropriate conversion function.
        """
        if self.vector_column_type in VECTOR_COLUMN_SQL_TYPES:
            return f"TO_{self.vector_column_type}({expr})"
        else:
            raise ValueError(f"Unsupported vector type: {self.vector_column_type}")

    def _convert_vector_embedding_to_column_type(self, expr: str) -> str:
        """
        Makes sure that an embedding produced by HANA's VECTOR_EMBEDDING
        aligns with the column type.

        Note that VECTOR_EMBEDDING always returns REAL_VECTORs.

        Args:
            expr (str): SQL expression producing an embedding vector.
        Returns:
            str: Wrapped expression if vector column type is not REAL_VECTOR,
                otherwise the original expression.
        """

        if "VECTOR_EMBEDDING" not in expr.upper():
            raise ValueError(f"Expected 'VECTOR_EMBEDDING' in '{expr}'")

        if self.vector_column_type != "REAL_VECTOR":
            return self._convert_to_target_vector_type(expr)

        return expr

    def create_hnsw_index(
        self,
        m: Optional[int] = None,  # Optional M parameter
        ef_construction: Optional[int] = None,  # Optional efConstruction parameter
        ef_search: Optional[int] = None,  # Optional efSearch parameter
        index_name: Optional[str] = None,  # Optional custom index name
    ) -> None:
        """
        Creates an HNSW vector index on a specified table and vector column with
        optional build and search configurations. If no configurations are provided,
        default parameters from the database are used. If provided values exceed the
        valid ranges, an error will be raised.
        The index is always created in ONLINE mode.

        Args:
            m: (Optional) Maximum number of neighbors per graph node
                (Valid Range: [4, 1000])
            ef_construction: (Optional) Maximal candidates to consider when building
                                the graph (Valid Range: [1, 100000])
            ef_search: (Optional) Minimum candidates for top-k-nearest neighbor
                                queries (Valid Range: [1, 100000])
            index_name: (Optional) Custom index name. Defaults to
                        <table_name>_<distance_strategy>_idx
        """
        # Set default index name if not provided
        distance_func_name = HANA_DISTANCE_FUNCTION[self.distance_strategy][0]
        default_index_name = f"{self.table_name}_{distance_func_name}_idx"
        # Use provided index_name or default
        index_name = (
            HanaDB._sanitize_name(index_name) if index_name else default_index_name
        )
        # Initialize build_config and search_config as empty dictionaries
        build_config = {}
        search_config = {}

        # Validate and add m parameter to build_config if provided
        if m is not None:
            m = HanaDB._sanitize_int(m)
            if not (4 <= m <= 1000):
                raise ValueError("M must be in the range [4, 1000]")
            build_config["M"] = m

        # Validate and add ef_construction to build_config if provided
        if ef_construction is not None:
            ef_construction = HanaDB._sanitize_int(ef_construction)
            if not (1 <= ef_construction <= 100000):
                raise ValueError("efConstruction must be in the range [1, 100000]")
            build_config["efConstruction"] = ef_construction

        # Validate and add ef_search to search_config if provided
        if ef_search is not None:
            ef_search = HanaDB._sanitize_int(ef_search)
            if not (1 <= ef_search <= 100000):
                raise ValueError("efSearch must be in the range [1, 100000]")
            search_config["efSearch"] = ef_search

        # Convert build_config and search_config to JSON strings if they contain values
        build_config_str = json.dumps(build_config) if build_config else ""
        search_config_str = json.dumps(search_config) if search_config else ""

        # Create the index SQL string with the ONLINE keyword
        sql_str = (
            f'CREATE HNSW VECTOR INDEX {index_name} ON "{self.table_name}" '
            f'("{self.vector_column}") '
            f"SIMILARITY FUNCTION {distance_func_name} "
        )

        # Append build_config to the SQL string if provided
        if build_config_str:
            sql_str += f"BUILD CONFIGURATION '{build_config_str}' "

        # Append search_config to the SQL string if provided
        if search_config_str:
            sql_str += f"SEARCH CONFIGURATION '{search_config_str}' "

        # Always add the ONLINE option
        sql_str += "ONLINE "
        cur = self.connection.cursor()
        try:
            cur.execute(sql_str)
        finally:
            cur.close()

    def add_texts(  # type: ignore[override]
        self,
        texts: Iterable[str],
        metadatas: Optional[list[dict]] = None,
        embeddings: Optional[list[list[float]]] = None,
        **kwargs: Any,
    ) -> list[str]:
        """Add texts to the vectorstore.

        Args:
            texts (Iterable[str]): Iterable of strings/text to add to the vectorstore.
            metadatas (Optional[list[dict]], optional): Optional list of metadatas.
                Defaults to None.
            embeddings (Optional[list[list[float]]], optional): Optional pre-generated
                embeddings. Defaults to None.

        Returns:
            list[str]: empty list
        """

        # decide how to add texts
        # using external embedding instance or internal embedding function of HanaDB
        if self.use_internal_embeddings:
            return self._add_texts_using_internal_embedding(
                texts, metadatas, embeddings
            )
        else:
            return self._add_texts_using_external_embedding(
                texts, metadatas, embeddings
            )

    def _add_texts_using_external_embedding(
        self,
        texts: Iterable[str],
        metadatas: Optional[list[dict]] = None,
        embeddings: Optional[list[list[float]]] = None,
        **kwargs: Any,
    ) -> list[str]:
        """Add texts using external embedding function"""
        # Create all embeddings of the texts beforehand to improve performance
        if embeddings is None:
            embeddings = self.embedding.embed_documents(list(texts))

        # Create sql parameters array
        sql_params = []
        for i, text in enumerate(texts):
            metadata = metadatas[i] if metadatas else {}
            metadata, extracted_special_metadata = self._split_off_special_metadata(
                metadata
            )
            sql_params.append(
                (
                    text,
                    json.dumps(HanaDB._sanitize_metadata_keys(metadata)),
                    self._serialize_binary_format(embeddings[i]),
                    *extracted_special_metadata,
                )
            )

        specific_metadata_columns_string = self._get_specific_metadata_columns_string()
        sql_str = (
            f'INSERT INTO "{self.table_name}" ("{self.content_column}", '
            f'"{self.metadata_column}", '
            f'"{self.vector_column}"{specific_metadata_columns_string}) '
            f"VALUES (?, ?, ?"
            f"{', ?' * len(self.specific_metadata_columns)});"
        )

        # Insert data into the table
        cur = self.connection.cursor()
        try:
            cur.executemany(sql_str, sql_params)
        finally:
            cur.close()
        return []

    def _add_texts_using_internal_embedding(
        self,
        texts: Iterable[str],
        metadatas: Optional[list[dict]] = None,
        embeddings: Optional[list[list[float]]] = None,
        **kwargs: Any,
    ) -> list[str]:
        """Add texts using internal embedding function"""
        sql_params = []
        for i, text in enumerate(texts):
            metadata = metadatas[i] if metadatas else {}
            metadata, extracted_special_metadata = self._split_off_special_metadata(
                metadata
            )
            parameters = [text]
            parameters.append(json.dumps(
                    HanaDB._sanitize_metadata_keys(metadata)
                ))  # Replace `metadata_value` with the actual value
            parameters.extend(self._generate_vector_embedding_sql_and_params(text, 'DOCUMENT')[1])
            # specific_metadata_values must align with the columns
            parameters.extend(extracted_special_metadata)
            sql_params.append(parameters)

        specific_metadata_str = ", ".join(
            f":{col}" for col in self.specific_metadata_columns
        )
        specific_metadata_columns_string = self._get_specific_metadata_columns_string()

        # Wrap VECTOR_EMBEDDING with vector type conversion if needed
        vector_embedding_sql = self._generate_vector_embedding_sql_and_params('text', 'DOCUMENT')[0]
        vector_embedding_sql = self._convert_vector_embedding_to_column_type(
            vector_embedding_sql
        )

        sql_str = (
            f'INSERT INTO "{self.table_name}" ("{self.content_column}", '
            f'"{self.metadata_column}", '
            f'"{self.vector_column}"{specific_metadata_columns_string}) '
            f"VALUES (?, ?, {vector_embedding_sql} "
            f"{(', ?'* len(self.specific_metadata_columns))});"
        )

        # Insert data into the table
        cur = self.connection.cursor()
        try:
            cur.executemany(sql_str, sql_params)
        finally:
            cur.close()
        return []

    def _get_specific_metadata_columns_string(self) -> str:
        """
        Helper function to generate the specific metadata columns as a SQL string.
        Returns:
            str: SQL string for specific metadata columns.
        """
        if not self.specific_metadata_columns:
            return ""
        return ', "' + '", "'.join(self.specific_metadata_columns) + '"'

    @classmethod
    def from_texts(  # type: ignore[no-untyped-def, override]
        cls: Type[HanaDB],
        texts: list[str],
        embedding: Embeddings,
        metadatas: Optional[list[dict]] = None,
        connection: dbapi.Connection = None,
        distance_strategy: DistanceStrategy = default_distance_strategy,
        table_name: str = default_table_name,
        content_column: str = default_content_column,
        metadata_column: str = default_metadata_column,
        vector_column: str = default_vector_column,
        vector_column_length: int = default_vector_column_length,
        vector_column_type: str = default_vector_column_type,
        *,
        specific_metadata_columns: Optional[list[str]] = None,
    ):
        """Create a HanaDB instance from raw documents.
        This is a user-friendly interface that:
            1. Embeds documents.
            2. Creates a table if it does not yet exist.
            3. Adds the documents to the table.
        This is intended to be a quick way to get started.
        """

        instance = cls(
            connection=connection,
            embedding=embedding,
            distance_strategy=distance_strategy,
            table_name=table_name,
            content_column=content_column,
            metadata_column=metadata_column,
            vector_column=vector_column,
            vector_column_length=vector_column_length,  # -1 means dynamic length
            vector_column_type=vector_column_type,
            specific_metadata_columns=specific_metadata_columns,
        )
        instance.add_texts(texts, metadatas)
        return instance

    def similarity_search(  # type: ignore[override]
        self, query: str, k: int = 4, filter: Optional[dict] = None
    ) -> list[Document]:
        """Return docs most similar to query.

        Args:
            query: Text to look up documents similar to.
            k: Number of Documents to return. Defaults to 4.
            filter: A dictionary of metadata fields and values to filter by.
                    Defaults to None.

        Returns:
            List of Documents most similar to the query
        """
        docs_and_scores = self.similarity_search_with_score(
            query=query, k=k, filter=filter
        )
        return [doc for doc, _ in docs_and_scores]

    def similarity_search_with_score(
        self, query: str, k: int = 4, filter: Optional[dict] = None
    ) -> list[tuple[Document, float]]:
        """Return documents and score values most similar to query.

        Args:
            query: Text to look up documents similar to.
            k: Number of Documents to return. Defaults to 4.
            filter: A dictionary of metadata fields and values to filter by.
                    Defaults to None.

        Returns:
            List of tuples (containing a Document and a score) that are
            most similar to the query
        """

        if self.use_internal_embeddings:
            # Internal embeddings: pass the query directly
            whole_result = self.similarity_search_with_score_and_vector_by_query(
                query=query, k=k, filter=filter
            )
        else:
            # External embeddings: generate embedding from the query
            embedding = self.embedding.embed_query(query)
            whole_result = self.similarity_search_with_score_and_vector_by_vector(
                embedding=embedding, k=k, filter=filter
            )

        return [(result_item[0], result_item[1]) for result_item in whole_result]

    def similarity_search_with_score_and_vector_by_vector(
        self, embedding: list[float], k: int = 4, filter: Optional[dict] = None
    ) -> list[tuple[Document, float, list[float]]]:
        """Return docs most similar to the given embedding.

        Args:
            embedding: Pre-computed embedding vector to search with.
            k: Number of Documents to return. Defaults to 4.
            filter: A dictionary of metadata fields and values to filter by.
                    Defaults to None.

        Returns:
            List of tuples, each containing:
            - Document: The matched document with its content and metadata
            - float: The similarity score
            - list[float]: The document's embedding vector
        """
        # Use the appropriate vector conversion function
        embedding_expr = self._convert_to_target_vector_type(expr=f"'{str(embedding)}'")

        return self._similarity_search_with_score_and_vector(
            embedding_expr, k=k, filter=filter
        )

    def similarity_search_with_score_and_vector_by_query(
        self, query: str, k: int = 4, filter: Optional[dict] = None
    ) -> list[tuple[Document, float, list[float]]]:
        """
        Return docs most similar to the given query.


        The query is vectorized in the database using HANA's internal
        embedding function. Therefore, the embedding instance provided
        during initialization must be of type HanaInternalEmbeddings.


        Args:
            query: Text to look up documents similar to.
            k: Number of Documents to return. Defaults to 4.
            filter: A dictionary of metadata fields and values to filter by.
                    Defaults to None.

        Returns:
            List of tuples, each containing:
            - Document: The matched document with its content and metadata
            - float: The similarity score
            - list[float]: The document's embedding vector
        """
        # Check if the embedding instance is of the correct type
        if not isinstance(self.embedding, HanaInternalEmbeddings):
            raise TypeError(
                "self.embedding must be an instance of "
                "HanaInternalEmbeddings to use the method "
                "similarity_search_with_score_and_vector_by_query"
            )

        vector_embedding_sql, vector_embedding_params = self._generate_vector_embedding_sql_and_params(query, 'QUERY')
        # Wrap VECTOR_EMBEDDING with vector type conversion if needed
        vector_embedding_sql = self._convert_vector_embedding_to_column_type(
            vector_embedding_sql
        )

        return self._similarity_search_with_score_and_vector(
            vector_embedding_sql,
            vector_embedding_params=vector_embedding_params,
            k=k,
            filter=filter,
        )

    def _similarity_search_with_score_and_vector(
        self,
        embedding_expr: str,
        k: int,
        filter: Optional[dict] = None,
        vector_embedding_params: Optional[dict] = None,
    ) -> list[tuple[Document, float, list[float]]]:
        """Perform similarity search and return documents with scores and vectors.

        Args:
            embedding_expr: SQL expression that generates or represents the embedding
                vector. For internal embeddings, this would be
                "VECTOR_EMBEDDING(?, 'QUERY', ?)",
                where ? will be replaced by parameters from query_params.
                For external embeddings, this would be "TO_REAL_VECTOR('...')".
            k: Number of documents to return. Defaults to 4.
            filter: Optional dictionary of metadata fields and values to filter by.
            query_params: Optional parameters for the embedding_expr SQL expression.
                For VECTOR_EMBEDDING function: [text, model_id] for the placeholders.
                For TO_REAL_VECTOR: None as the vector is included in the expression.

        Returns:
            List of tuples, each containing:
            - Document: The matched document with its content and metadata
            - float: The similarity score
            - list[float]: The document's embedding vector
        """

        _validate_k(k)

        result = []
        distance_func_name = HANA_DISTANCE_FUNCTION[self.distance_strategy][0]

        # Generate metadata projection for filtered results
        projected_metadata_columns = self._extract_keyword_search_columns(filter)
        metadata_projection = ""
        if projected_metadata_columns:
            metadata_projection = self._create_metadata_projection(
                projected_metadata_columns
            )
        if metadata_projection:
            from_clause = INTERMEDIATE_TABLE_NAME
        else:
            from_clause = f'"{self.table_name}"'

        sql_str = (
            f"{metadata_projection} "
            f"SELECT TOP {k}"
            f'  "{self.content_column}", '  # row[0]
            f'  "{self.metadata_column}", '  # row[1]
            f'  "{self.vector_column}", '  # row[2]
            f'  {distance_func_name}("{self.vector_column}", '
            f"  {embedding_expr}) AS CS "  # row[3]
            f"FROM {from_clause}"
        )
        parameters = []
        if vector_embedding_params:
            parameters += vector_embedding_params
        where_clause, where_parameters = CreateWhereClause(self)(filter)
        if where_clause:
            sql_str += f" {where_clause}"
            parameters += where_parameters
        sql_str += f" order by CS {HANA_DISTANCE_FUNCTION[self.distance_strategy][1]}"

        try:
            cur = self.connection.cursor()
            if parameters:
                cur.execute(sql_str, parameters)
            else:
                cur.execute(sql_str)
            if cur.has_result_set():
                rows = cur.fetchall()
                for row in rows:
                    js = json.loads(row[1])
                    doc = Document(page_content=row[0], metadata=js)
                    result_vector = self._handle_vector_output_type(row[2])
                    result.append((doc, row[3], result_vector))
        except dbapi.Error:
            logger.error(f"SQL Statement: {sql_str}")
            logger.error(f"Parameters: {parameters}")
            raise
        finally:
            cur.close()
        return result

    def _extract_keyword_search_columns(
        self, filter: Optional[dict] = None
    ) -> list[str]:
        """
        Extract metadata columns used with `$contains` in the filter.
        Scans the filter to find unspecific metadata columns used
        with the `$contains` operator.
        Args:
            filter: A dictionary of filter criteria.
        Returns:
            list of metadata column names for keyword searches.
        Example:
            filter = {"$or": [
                {"title": {"$contains": "barbie"}},
                {"VEC_TEXT": {"$contains": "fred"}}]}
            Result: ["title"]
        """
        keyword_columns: set[str] = set()
        self._recurse_filters(keyword_columns, filter)
        return list(keyword_columns)

    def _recurse_filters(
        self,
        keyword_columns: set[str],
        filter_obj: Optional[dict[Any, Any]],
        parent_key: Optional[str] = None,
    ) -> None:
        """
        Recursively process the filter dictionary
        to find metadata columns used with `$contains`.
        """
        if isinstance(filter_obj, dict):
            for key, value in filter_obj.items():
                if key == CONTAINS_OPERATOR:
                    # Add the parent key as it's the metadata column being filtered
                    if parent_key and not (
                        parent_key == self.content_column
                        or parent_key in self.specific_metadata_columns
                    ):
                        keyword_columns.add(parent_key)
                elif key in LOGICAL_OPERATORS_TO_SQL:  # Handle logical operators
                    for subfilter in value:
                        self._recurse_filters(keyword_columns, subfilter)
                else:
                    self._recurse_filters(keyword_columns, value, parent_key=key)

    def _create_metadata_projection(self, projected_metadata_columns: list[str]) -> str:
        """
        Generate a SQL `WITH` clause to project metadata columns for keyword search.
        Args:
            projected_metadata_columns: List of metadata column names for projection.
        Returns:
            A SQL `WITH` clause string.
        Example:
            Input: ["title", "author"]
            Output:
            WITH intermediate_result AS (
                SELECT *,
                JSON_VALUE(metadata_column, '$.title') AS "title",
                JSON_VALUE(metadata_column, '$.author') AS "author"
                FROM "table_name"
            )
        """

        metadata_columns = [
            (
                f"JSON_VALUE({self.metadata_column}, '$.{HanaDB._sanitize_name(col)}') "
                f'AS "{HanaDB._sanitize_name(col)}"'
            )
            for col in projected_metadata_columns
        ]
        return (
            f"WITH {INTERMEDIATE_TABLE_NAME} AS ("
            f"SELECT *, {', '.join(metadata_columns)} "
            f"FROM \"{self.table_name}\")"
        )

    def similarity_search_by_vector(  # type: ignore[override]
        self, embedding: list[float], k: int = 4, filter: Optional[dict] = None
    ) -> list[Document]:
        """Return docs most similar to embedding vector.

        Args:
            embedding: Embedding to look up documents similar to.
            k: Number of Documents to return. Defaults to 4.
            filter: A dictionary of metadata fields and values to filter by.
                    Defaults to None.

        Returns:
            List of Documents most similar to the query vector.
        """
        whole_result = self.similarity_search_with_score_and_vector_by_vector(
            embedding=embedding, k=k, filter=filter
        )
        docs_and_scores = [
            (result_item[0], result_item[1]) for result_item in whole_result
        ]
        return [doc for doc, _ in docs_and_scores]

    def delete(  # type: ignore[override]
        self, ids: Optional[list[str]] = None, filter: Optional[dict] = None
    ) -> Optional[bool]:
        """Delete entries by filter with metadata values

        Args:
            ids: Deletion with ids is not supported! A ValueError will be raised.
            filter: A dictionary of metadata fields and values to filter by.
                    An empty filter ({}) will delete all entries in the table.

        Returns:
            Optional[bool]: True, if deletion is technically successful.
            Deletion of zero entries, due to non-matching filters is a success.
        """

        if ids is not None:
            raise ValueError("Deletion via ids is not supported")

        if filter is None:
            raise ValueError("Parameter 'filter' is required when calling 'delete'")

        where_clause, parameters = CreateWhereClause(self)(filter)
        sql_str = f'DELETE FROM "{self.table_name}" {where_clause}'

        try:
            cur = self.connection.cursor()
            cur.execute(sql_str, parameters)
        finally:
            cur.close()

        return True

    async def adelete(  # type: ignore[override]
        self, ids: Optional[list[str]] = None, filter: Optional[dict] = None
    ) -> Optional[bool]:
        """Delete by vector ID or other criteria.

        Args:
            ids: List of ids to delete.

        Returns:
            Optional[bool]: True if deletion is successful,
            False otherwise, None if not implemented.
        """
        return await run_in_executor(None, self.delete, ids=ids, filter=filter)

    def _embed_query_hana_internal(self, query: str) -> list[float]:
        """
        Generates query embedding using HANA's internal embedding engine.
        """
        vector_embedding_sql, vector_embedding_sql_params = self._generate_vector_embedding_sql_and_params(query, 'QUERY')
        vector_embedding_sql = self._convert_vector_embedding_to_column_type(
            vector_embedding_sql
        )
        sql_str = f"SELECT {vector_embedding_sql} FROM sys.DUMMY;"
        cur = self.connection.cursor()
        try:
            cur.execute(
                sql_str,
                vector_embedding_sql_params,
            )
            if cur.has_result_set():
                res = cur.fetchall()
                return self._handle_vector_output_type(res[0][0])
            else:
                raise ValueError("No result set returned for query embedding.")
        finally:
            cur.close()

    def max_marginal_relevance_search(  # type: ignore[override]
        self,
        query: str,
        k: int = 4,
        fetch_k: int = 20,
        lambda_mult: float = 0.5,
        filter: Optional[dict] = None,
    ) -> list[Document]:
        """Return docs selected using the maximal marginal relevance.

        Maximal marginal relevance optimizes for similarity to query AND diversity
        among selected documents.

        Args:
            query: search query text.
            k: Number of Documents to return. Defaults to 4.
            fetch_k: Number of Documents to fetch to pass to MMR algorithm.
            lambda_mult: Number between 0 and 1 that determines the degree
                        of diversity among the results with 0 corresponding
                        to maximum diversity and 1 to minimum diversity.
                        Defaults to 0.5.
            filter: Filter on metadata properties, e.g.
                            {
                                "str_property": "foo",
                                "int_property": 123
                            }
        Returns:
            List of Documents selected by maximal marginal relevance.
        """
        if not self.use_internal_embeddings:
            embedding = self.embedding.embed_query(query)
        else:  # generates embedding using the internal embedding function of HanaDb
            embedding = self._embed_query_hana_internal(query)

        return self.max_marginal_relevance_search_by_vector(
            embedding=embedding,
            k=k,
            fetch_k=fetch_k,
            lambda_mult=lambda_mult,
            filter=filter,
        )

    def _parse_float_array_from_string(array_as_string: str) -> list[float]:  # type: ignore[misc]
        array_wo_brackets = array_as_string[1:-1]
        return [float(x) for x in array_wo_brackets.split(",")]

    def max_marginal_relevance_search_by_vector(  # type: ignore[override]
        self,
        embedding: list[float],
        k: int = 4,
        fetch_k: int = 20,
        lambda_mult: float = 0.5,
        filter: Optional[dict] = None,
    ) -> list[Document]:
        _validate_k_and_fetch_k(k, fetch_k)

        whole_result = self.similarity_search_with_score_and_vector_by_vector(
            embedding=embedding, k=fetch_k, filter=filter
        )
        embeddings = [result_item[2] for result_item in whole_result]
        mmr_doc_indexes = maximal_marginal_relevance(
            np.array(embedding), embeddings, lambda_mult=lambda_mult, k=k
        )

        return [whole_result[i][0] for i in mmr_doc_indexes]

    async def amax_marginal_relevance_search_by_vector(  # type: ignore[override]
        self,
        embedding: list[float],
        k: int = 4,
        fetch_k: int = 20,
        lambda_mult: float = 0.5,
    ) -> list[Document]:
        """Return docs selected using the maximal marginal relevance."""
        return await run_in_executor(
            None,
            self.max_marginal_relevance_search_by_vector,
            embedding=embedding,
            k=k,
            fetch_k=fetch_k,
            lambda_mult=lambda_mult,
        )

    @staticmethod
    def _cosine_relevance_score_fn(distance: float) -> float:
        return distance

    def _select_relevance_score_fn(self) -> Callable[[float], float]:
        """
        The 'correct' relevance function
        may differ depending on a few things, including:
        - the distance / similarity metric used by the VectorStore
        - the scale of your embeddings (OpenAI's are unit normed. Many others are not!)
        - embedding dimensionality
        - etc.

        Vectorstores should define their own selection based method of relevance.
        """
        if self.distance_strategy == DistanceStrategy.COSINE:
            return HanaDB._cosine_relevance_score_fn
        elif self.distance_strategy == DistanceStrategy.EUCLIDEAN_DISTANCE:
            return HanaDB._euclidean_relevance_score_fn
        else:
            raise ValueError(
                "Unsupported distance_strategy: {}".format(self.distance_strategy)
            )



from enum import Enum


class DistanceStrategy(str, Enum):
    """Enumerator of the Distance strategies for calculating distances
    between vectors."""

    EUCLIDEAN_DISTANCE = "EUCLIDEAN_DISTANCE"
    MAX_INNER_PRODUCT = "MAX_INNER_PRODUCT"
    DOT_PRODUCT = "DOT_PRODUCT"
    JACCARD = "JACCARD"
    COSINE = "COSINE"


def _validate_k(k: int):
    if not isinstance(k, int) or k <= 0:
        raise ValueError("Parameter 'k' must be an integer greater than 0")


def _validate_k_and_fetch_k(k: int, fetch_k: int):
    _validate_k(k)
    if not isinstance(fetch_k, int) or fetch_k < k:
        raise ValueError(
            "Parameter 'fetch_k' must be an integer greater than or equal to 'k'"
        )



from importlib import metadata

from langchain_hana.chains import HanaSparqlQAChain
from langchain_hana.embeddings import HanaInternalEmbeddings
from langchain_hana.graphs import HanaRdfGraph
from langchain_hana.vectorstores import HanaDB
from langchain_hana.structured_query import HanaTranslator

try:
    __version__ = metadata.version(__package__)
except metadata.PackageNotFoundError:
    # Case where package metadata is not available.
    __version__ = ""
del metadata  # optional, avoids polluting the results of dir(__package__)

__all__ = [
    "HanaDB",
    "HanaTranslator",
    "HanaInternalEmbeddings",
    "HanaRdfGraph",
    "HanaSparqlQAChain",
    "__version__",
]



import logging
from typing import List, Tuple, Any

logger = logging.getLogger(__name__)

COMPARISONS_TO_SQL = {
    "$eq": "=",
    "$ne": "<>",
    "$lt": "<",
    "$lte": "<=",
    "$gt": ">",
    "$gte": ">=",
}

IN_OPERATORS_TO_SQL = {
    "$in": "IN",
    "$nin": "NOT IN",
}

BETWEEN_OPERATOR = "$between"

LIKE_OPERATOR = "$like"

CONTAINS_OPERATOR = "$contains"


class ContainsNeedsSpecialSqlSyntax:
    def __repr__(self) -> None:  # type: ignore[override]
        raise AssertionError(f"{CONTAINS_OPERATOR} needs special SQL syntax")


COLUMN_OPERATORS = {
    **COMPARISONS_TO_SQL,
    **IN_OPERATORS_TO_SQL,
    BETWEEN_OPERATOR: "BETWEEN",
    LIKE_OPERATOR: "LIKE",
    CONTAINS_OPERATOR: ContainsNeedsSpecialSqlSyntax(),
}

LOGICAL_OPERATORS_TO_SQL = {"$and": "AND", "$or": "OR"}

class CreateWhereClause:
    def __init__(self, hanaDb: Any) -> None:
        self.specific_metadata_columns = hanaDb.specific_metadata_columns
        self.metadata_column = hanaDb.metadata_column

    def __call__(self, filter):  # type: ignore[no-untyped-def]
        """Serializes filter to a where clause (prepared_statement) and parameters

        The where clause should be appended to an existing SQL statement.

        Example usage:
        where_clause, parameters = CreateWhereClause(hanaDb)(filter)
        cursor.execute(f"{stmt} {where_clause}", parameters)
        """
        if filter:
            statement, parameters = self._create_where_clause(filter)
            assert statement.count("?") == len(parameters)
            return f"WHERE {statement}", parameters
        else:
            return "", []

    def _create_where_clause(self, filter: dict) -> Tuple[str, List]:
        if not filter:
            raise ValueError("Empty filter")
        statements = []
        parameters = []
        for key, value in filter.items():
            if key.startswith("$"):
                # Generic filter objects may only have logical operators.
                if key not in LOGICAL_OPERATORS_TO_SQL:
                    raise ValueError(f"Unexpected operator {key=} in {filter=}")
                ret_sql_clause, ret_query_tuple = self._sql_serialize_logical_operation(
                    key, value
                )
            else:
                if not isinstance(value, (bool, int, str, dict)):
                    raise ValueError(f"Unsupported filter value type: {type(value)}")
                if isinstance(value, dict) and "type" not in value:
                    # Value is an operator.
                    if len(value) != 1:
                        raise ValueError(
                            "Expecting a single entry 'operator: operands'"
                            f", but got {value=}"
                        )
                    operator, operands = list(value.items())[0]
                    ret_sql_clause, ret_query_tuple = (
                        self._sql_serialize_column_operation(key, operator, operands)
                    )
                else:
                    placeholder, value = (
                        CreateWhereClause._determine_typed_sql_placeholder(value)
                    )
                    ret_sql_clause = f"{self._create_selector(key)} = {placeholder}"
                    ret_query_tuple = [value]
            statements.append(ret_sql_clause)
            parameters += ret_query_tuple
        return CreateWhereClause._sql_serialize_logical_clauses(
            "AND", statements
        ), parameters

    def _sql_serialize_column_operation(
        self, column: str, operator: str, operands: any
    ) -> Tuple[str, List]:
        if operator in LOGICAL_OPERATORS_TO_SQL:
            raise ValueError(
                f"Did not expect oerator from {LOGICAL_OPERATORS_TO_SQL=}"
                f", but got {operator=}"
            )
        if operator not in COLUMN_OPERATORS:
            raise ValueError(f"{operator=} not in {COLUMN_OPERATORS.keys()=}")
        if operator == CONTAINS_OPERATOR:
            placeholder, value = CreateWhereClause._determine_typed_sql_placeholder(
                operands
            )
            statement = (
                f"SCORE({placeholder} IN (\"{column}\" EXACT SEARCH MODE 'text')) > 0"
            )
            return statement, [value]
        sql_operator = COLUMN_OPERATORS[operator]
        selector = self._create_selector(column)
        if operator == BETWEEN_OPERATOR:
            if len(operands) != 2:
                raise ValueError(f"Expected 2 operands, but got {operands=}")
            from_placeholder, from_value = (
                CreateWhereClause._determine_typed_sql_placeholder(operands[0])
            )
            to_placeholder, to_value = (
                CreateWhereClause._determine_typed_sql_placeholder(operands[1])
            )
            statement = (
                f"{selector} {sql_operator} {from_placeholder} AND {to_placeholder}"
            )
            return statement, [from_value, to_value]
        if operator in IN_OPERATORS_TO_SQL:
            if not isinstance(operands, list):
                raise ValueError(f"Expected a list, but got {operands=}")
            placeholder_value_list = [
                CreateWhereClause._determine_typed_sql_placeholder(item)
                for item in operands
            ]
            placeholders = ", ".join([item[0] for item in placeholder_value_list])
            values = [item[1] for item in placeholder_value_list]
            statement = f"{selector} {sql_operator} ({placeholders})"
            return statement, values
        # Default behavior for single value operators.
        placeholder, value = CreateWhereClause._determine_typed_sql_placeholder(
            operands
        )
        statement = f"{selector} {sql_operator} {placeholder}"
        return statement, [value]

    @staticmethod
    def _determine_typed_sql_placeholder(value):  # type: ignore[no-untyped-def]

        the_type = type(value)

        # Handle plain values.
        if the_type is bool:
            return "TO_BOOLEAN(?)", "true" if value else "false"
        if the_type in (int, float):
            return "TO_DOUBLE(?)", value
        
        # Do not accept empty values.
        if not value:
            raise ValueError("No operands provided")

        # Handle container types: only allowed for dates.
        if isinstance(value, dict) and ("type" in value) and (value["type"] == "date"):
            return "TO_DATE(?)", value["date"]
        if isinstance(value, (dict, list, tuple)):
            raise ValueError(f"Cannot handle {value=}")
        
        logger.warning(f"Plain SQL Placeholder '?' for {value=}")
        return "?", value

    @staticmethod
    def _sql_serialize_logical_clauses(
        sql_operator: str, sql_clauses: list[str]
    ) -> str:
        supported_operators = LOGICAL_OPERATORS_TO_SQL.values()
        if sql_operator not in supported_operators:
            raise ValueError(f"{sql_operator=}, is not in {supported_operators=}")
        if not sql_clauses:
            raise ValueError("sql_clauses is empty")
        if not all(sql_clauses):
            raise ValueError(f"Empty sql clause in {sql_clauses=}")
        if len(sql_clauses) == 1:
            return sql_clauses[0]
        return f" {sql_operator} ".join([f"({clause})" for clause in sql_clauses])

    def _sql_serialize_logical_operation(
        self, operator: str, operands: List
    ) -> Tuple[str, List]:
        if not isinstance(operands, list):
            raise ValueError(f"Unexpected operands for {operator=}: {operands=}")
        if operator not in LOGICAL_OPERATORS_TO_SQL:
            raise ValueError(
                f"Expected operator from {LOGICAL_OPERATORS_TO_SQL=}"
                f", but got {operator=}"
            )
        sql_clauses, query_tuple = [], []
        for operand in operands:
            ret_sql_clause, ret_query_tuple = self._create_where_clause(operand)
            sql_clauses.append(ret_sql_clause)
            query_tuple += ret_query_tuple
        return (
            CreateWhereClause._sql_serialize_logical_clauses(
                LOGICAL_OPERATORS_TO_SQL[operator], sql_clauses
            ),
            query_tuple,
        )

    def _create_selector(self, column: str) -> str:
        if column in self.specific_metadata_columns:
            return f'"{column}"'
        else:
            return f"JSON_VALUE({self.metadata_column}, '$.{column}')"



from langchain_core.embeddings import Embeddings


class HanaInternalEmbeddings(Embeddings):
    """
    A dummy embeddings class for use with HANA's internal embedding functionality.
    This class prevents the use of standard embedding methods and ensures that
    internal embeddings are handled exclusively via database queries.
    """

    def __init__(self, internal_embedding_model_id: str, remote_source: str = ""):
        """
        Initialize the HanaInternalEmbeddings instance.
        Args:
            internal_embedding_model_id (str): The ID of the internal embedding model
                                               used by the HANA database.
        """
        self.model_id = internal_embedding_model_id
        self.remote_source = remote_source

    def embed_query(self, text: str) -> list[float]:
        """
        Override the embed_query method to raise an error.
        This method is not applicable for HANA internal embeddings.
        Raises:
            NotImplementedError: Always raised to show that this method is unsupported.
        """
        raise NotImplementedError(
            "Internal embeddings cannot be used externally. "
            "Use HANA's internal embedding functionality instead."
        )

    def embed_documents(self, texts: list[str]) -> list[list[float]]:
        """
        Override the embed_documents method to raise an error.
        This method is not applicable for HANA internal embeddings.
        Raises:
            NotImplementedError: Always raised to show that this method is unsupported.
        """
        raise NotImplementedError(
            "Internal embeddings cannot be used externally. "
            "Use HANA's internal embedding functionality instead."
        )

    def get_model_id(self) -> str:
        """
        Retrieve the internal embedding model ID.
        Returns:
            str: The ID of the internal embedding model.
        """
        return self.model_id
    
    def get_remote_source(self) -> str:
        """
        Retrieve the remote source name.
        Returns:
            str: The remote source associated with the internal embedding.
        """
        return self.remote_source
